'use strict';

const Utils = require('./utils');
const Constants = require('./constants');
const Crypto = require('crypto');

const checkInternalInvocaion = (methodName, userRoles) => {
  if (!userRoles.includes(Constants.SERVER_CODE_USER_ROLE)) {
    throw new Error(`Invoke "Twilio.${ methodName }" method is allowed from BusinessLogic only.`);
  }
};

class Twilio {

  /**
   * @param {String} phoneNumber
   * @param {String} message
   * @param {String} mediaUrl
   * @param {Object} options
   *
   * @returns {Object}
   **/
  sendMessage(phoneNumber, message, mediaUrl, options) {
    options = options || {};

    const { accountSID, authToken, statusCallback } = this.config;
    const senderPhoneNumber = options.senderPhoneNumber || this.config.senderPhoneNumber;

    if (!accountSID) {
      throw new Error('"Twilio Account SID" is not configured.');
    }

    if (!authToken) {
      throw new Error('"Twilio Auth Token" is not configured.');
    }

    if (!senderPhoneNumber) {
      throw new Error('Configure "Twilio Sender Phone Number" or pass "senderPhoneNumber" argument to the method.');
    }

    return Utils.sendMessage(accountSID, authToken, {
      phoneNumber,
      message,
      mediaUrl,
      senderPhoneNumber,
      statusCallback
    });
  }

  /**
   * @param {String} phoneNumber
   *
   * @returns {Object} transaction
   **/
  async sendVerificationCode(phoneNumber) {
    checkInternalInvocaion('sendVerificationCode', this.request.context.userRoles);

    return this._sendVerificationCode(phoneNumber)
  }

  /**
   * @param {String} transactionId
   * @param {String} code
   *
   * @returns {Object} transaction
   **/
  async checkVerificationCode(transactionId, code) {
    checkInternalInvocaion('checkVerificationCode', this.request.context.userRoles);

    return this._checkVerificationCode(transactionId, code)
  }

  /**
   * @param {String} transactionId
   * @param {String} code
   *
   * @returns {Object}
   **/
  async loginWithCode(transactionId, code) {
    const { phoneNumberField } = this.config;

    if (!phoneNumberField) {
      throw new Error('Configure "User Phone Number Field"');
    }

    const { phoneNumber } = await this._checkVerificationCode(transactionId, code);

    const user = await this._findUserByPhoneNumber(phoneNumber);

    if (!user) {
      throw new Error('TransactionId is invalid');
    }

    return Backendless.Users.login(user.objectId)
  }

  /**
   * @param {String} phoneNumber
   *
   * @returns {Object}
   **/
  async loginWithPhoneNumber(phoneNumber) {
    const { phoneNumberField } = this.config;

    if (!phoneNumberField) {
      throw new Error('Configure "User Phone Number Field"');
    }

    const user = await this._findUserByPhoneNumber(phoneNumber);

    if (!user) {
      throw new Error('There is no user with provided phone number');
    }

    const transactionId = await this._sendVerificationCode(user[phoneNumberField]);

    return {
      transactionId
    }
  }

  /**
   * @param {String} phoneNumber
   * @param {String} password
   *
   * @returns {Object}
   **/
  async loginWithPhoneNumberAndPassword(phoneNumber, password) {
    const { phoneNumberField } = this.config;

    if (!phoneNumberField) {
      throw new Error('Configure "User Phone Number Field"');
    }

    const user = await this._findUserByPhoneNumber(phoneNumber);

    if (!user) {
      throw new Error('There is no user with provided phone number');
    }

    const identityName = await this._getUserIdentityName();

    await Backendless.Users.login(user[identityName], password);

    const transactionId = await this._sendVerificationCode(user[phoneNumberField]);

    return {
      transactionId
    }
  }

  /**
   * @route PUT /enableSMSCodeAuth
   * @returns {Promise<Object>}
   */
  async enable2FA() {
    if (!this.request.context.userToken) {
      throw new Error('You must be authenticated to use this endpoint')
    }

    const { userId } = this.request.context

    const user = await Backendless.Data.of('Users').findById(userId)

    if (user.smsCodeAuth) {
      throw new Error('User has already enabled 2FA')
    }

    if (!user.phoneNumber) {
      throw new Error('User should have phone number')
    }

    await Backendless.Data.of('Users').save({ objectId: userId, smsCodeAuth: true })

    const transactionId = await Backendless.CustomServices.invoke('Twilio', 'sendVerificationCode', user.phoneNumber)

    if (transactionId) {
      return { transactionId }
    }
  }

  /**
   * @route PUT /disableSMSCodeAuth
   * @returns {Promise<void>}
   */
  async disable2FA() {
    if (!this.request.context.userToken) {
      throw new Error('You must be authenticated to use this endpoint')
    }

    const { userId } = this.request.context

    const user = await Backendless.Data.of('Users').findById(userId)

    if (!user.smsCodeAuth) {
      throw new Error('User has already disabled 2FA')
    }

    await Backendless.Data.of('Users').save({ objectId: userId, smsCodeAuth: false })
  }

  /**
   * @private
   **/
  async _getUserIdentityName() {
    const properties = await Backendless.Users.describeUserClass();

    for (const property of properties) {
      if (property.identity) {
        return property.name
      }
    }
  }

  /**
   * @private
   **/
  async _findUserByPhoneNumber(phoneNumber) {
    const { phoneNumberField } = this.config;

    const query = Backendless.Data.QueryBuilder.create()
      .setWhereClause(`${ phoneNumberField } = '${ phoneNumber }'`)

    const users = await Backendless.Data.of(Backendless.User).find(query);

    return users[0]
  }

  /**
   * @private
   **/
  _generateTransactionUid(transactionId) {
    const { accountSID, authToken } = this.config;

    return Crypto
      .createHmac('sha256', `${ accountSID }-${ authToken }`)
      .update(transactionId)
      .digest('hex');
  }

  /**
   * @private
   **/
  async _sendVerificationCode(phoneNumber) {
    const { senderPhoneNumber, verificationCodeSMSTemplate } = this.config;

    if (!senderPhoneNumber) {
      throw new Error(`Configure "Twilio Sender Phone Number"`);
    }

    if (!verificationCodeSMSTemplate) {
      throw new Error('Configure "Verification Code SMS Template"');
    }

    if (verificationCodeSMSTemplate.indexOf(Constants.CODE_TEMPLATE_TOKEN) === -1) {
      throw new Error(`"Verification Code SMS Template" config item must contains: ${ Constants.CODE_TEMPLATE_TOKEN } .`);
    }

    if (!phoneNumber) {
      throw new Error(`Phone Number is required argument`);
    }

    const code = Utils.generateVerificationCode();

    const sms = verificationCodeSMSTemplate.replace(Constants.CODE_TEMPLATE_TOKEN, code);

    const { sid } = await this.sendMessage(phoneNumber, sms, undefined, { senderPhoneNumber });

    const transactionUid = this._generateTransactionUid(sid);
    const transaction = Utils.encrypt({ code, phoneNumber }, transactionUid);

    await Backendless.Cache.put(transactionUid, transaction, Constants.VERIFICATION_CODE_TTL);

    return sid
  }

  /**
   * @private
   **/
  async _checkVerificationCode(transactionId, code) {
    const transactionUid = this._generateTransactionUid(transactionId);
    const transaction = Utils.decrypt(await Backendless.Cache.get(transactionUid), transactionUid);

    if (!transaction) {
      throw new Error('Transaction is not valid');
    }

    if (code !== transaction.code) {
      throw new Error('Verification code is not valid');
    }

    await Backendless.Cache.remove(transactionUid);

    return transaction
  }

}

Backendless.ServerCode.addService(Twilio, [
  {
    order      : 0,
    displayName: 'Twilio Account SID',
    name       : 'accountSID',
    type       : Backendless.ServerCode.ConfigItems.TYPES.STRING,
    required   : true,
    hint       : 'Please specify your Twilio Account SID'
  },
  {
    order      : 1,
    displayName: 'Twilio Auth Token',
    name       : 'authToken',
    type       : Backendless.ServerCode.ConfigItems.TYPES.STRING,
    required   : true,
    hint       : 'Please specify your Twilio Auth Token'
  },
  {
    order      : 2,
    displayName: 'Twilio Sender Phone Number',
    name       : 'senderPhoneNumber',
    type       : Backendless.ServerCode.ConfigItems.TYPES.STRING,
    required   : true,
    hint       : 'Please specify your Twilio Sender Phone Number'
  },
  {
    order       : 3,
    displayName : 'Verification Code SMS Template',
    name        : 'verificationCodeSMSTemplate',
    type        : Backendless.ServerCode.ConfigItems.TYPES.STRING,
    defaultValue: `Your Verification Code is: ${ Constants.CODE_TEMPLATE_TOKEN }.`,
    required    : true,
    hint        : 'Please specify Verification SMS Template'
  },
  {
    order       : 4,
    displayName : 'User Phone Number Field',
    name        : 'phoneNumberField',
    type        : Backendless.ServerCode.ConfigItems.TYPES.STRING,
    defaultValue: 'phoneNumber',
    required    : true,
    hint        : 'Please specify User Phone Number Field'
  },
  {
    order       : 5,
    displayName : 'Message Status Callback URL',
    name        : 'statusCallback',
    type        : Backendless.ServerCode.ConfigItems.TYPES.STRING,
    defaultValue: '',
    required    : false
  }
]);
